"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from typing import Any
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import Times

if TYPE_CHECKING:
    from .params import UbiqClientParams
    from .router import UbiqRouter  # noqa: F401
    from .router import _FETCH  # noqa: F401
    from ..homie import Homie



_SOURCES = dict[str, '_FETCH']
_ROUTERS = dict[str, 'UbiqRouter']



class UbiqClient:
    """
    Contain the relevant attributes about the related device.

    :param homie: Primary class instance for Homie Automate.
    :param name: Name of the object within the Homie config.
    """

    __homie: 'Homie'
    __params: 'UbiqClientParams'

    __name: str

    __sources: Optional[_SOURCES]
    __routers: Optional[_ROUTERS]


    def __init__(
        self,
        homie: 'Homie',
        name: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        homie.log_d(
            base='UbiqClient',
            name=name,
            status='initial')


        clients = (
            homie.params
            .ubiq_clients)

        assert clients is not None

        params = clients[name]


        self.__homie = homie
        self.__params = params
        self.__name = name
        self.__sources = None
        self.__routers = None


        self.__validate_params()

        homie.log_d(
            base='UbiqClient',
            name=name,
            status='created')


    def __validate_params(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        routers = self.homie.ubiq_routers
        router_name = self.params.router

        assert routers is not None
        assert (
            router_name in routers
            or router_name is None)


    def refresh_source(
        self,
    ) -> None:
        """
        Update the class instance from cached upstream response.
        """

        homie = self.homie
        routers = homie.ubiq_routers
        params = self.params

        assert routers is not None

        _sources: _SOURCES = {}
        _routers: _ROUTERS = {}


        expect = params.router

        for router in routers.values():

            name = router.name

            if (expect is not None
                    and expect != name):
                continue


            found = router.get_source(
                ubid=(
                    params.ubid
                    or params.mac
                    or params.ip),
                label=params.label)

            if found is None:
                continue


            _sources[name] = found
            _routers[name] = router


        self.__sources = _sources or None
        self.__routers = _routers or None


    @property
    def homie(
        self,
    ) -> 'Homie':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.__homie


    @property
    def params(
        self,
    ) -> 'UbiqClientParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def name(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__name


    @property
    def routers(
        self,
    ) -> Optional[_ROUTERS]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        if self.__routers is None:
            return None

        return dict(self.__routers)


    @property
    def sources(
        self,
    ) -> Optional[_SOURCES]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        if self.__sources is None:
            return None

        return deepcopy(self.__sources)


    @property
    def uniques(
        self,
    ) -> Optional[dict[str, str]]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        sources = self.__sources

        if sources is None:
            return None

        ubid: dict[str, str] = {}

        items = sources.items()

        for key, value in items:
            ubid[key] = value['_id']

        return ubid


    @property
    def present(
        self,
    ) -> list[str] | Literal[False]:
        """
        Return the list of routers which the client is present.

        :returns: List of routers which the client is present,
            or `False` boolean when not present on any routers.
        """

        return (
            list(self.__sources)
            if self.__sources is not None
            else False)


    @property
    def connect(
        self,
    ) -> list[str] | Literal[False]:
        """
        Return the list of routers which the client is connected.

        :returns: List of routers which the client is connected,
            or `False` boolean when not connected to any routers.
        """

        routers: list[str] = []

        sources = self.__sources

        if sources is None:
            return False


        items = sources.items()

        for name, source in items:

            connect = False

            if '_uptime_by_ugw' in source:
                connect = True

            if '_uptime_by_uap' in source:
                connect = True

            if connect is True:
                routers.append(name)


        return routers or False


    @property
    def lastseen(
        self,
    ) -> Optional[dict[str, Times]]:
        """
        Return the last time that the device was seen by router.

        :returns: Last time that the device was seen by router.
        """

        returned: dict[str, Times] = {}

        sources = self.__sources

        if sources is None:
            return None


        items1 = sources.items()

        for name, source in items1:

            lseen = max(
                source['last_seen'])

            returned |= {
                name: Times(lseen)}


        times: Optional[Times] = None


        items2 = returned.items()

        for name, _times in items2:

            if (times is None
                    or _times > times):
                times = _times

        if times is not None:
            returned['_latest'] = times


        return returned or None


    def homie_dumper(
        self,
    ) -> dict[str, Any]:
        """
        Return the content related to the project dumper script.

        :returns: Content related to the project dumper script.
        """

        params = (
            self.params.model_dump())

        self.refresh_source()

        return {
            'name': self.name,
            'uniques': self.uniques,
            'present': self.present,
            'connect': self.connect,
            'params': params}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import Optional

from encommon.times import Times

from pydantic import BaseModel



class WhenTimePeriodParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    :param start: Determine the start for the desired period.
    :param stop: Determine the ending for the desired period.
    :param tzname: Name of the timezone associated to period.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    start: Optional[str] = None
    stop: Optional[str] = None
    tzname: str = 'UTC'


    def __init__(
        self,
        **data: Any,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        start = data.get('start')
        stop = data.get('stop')
        tzname = data.get('tzname')


        if start is not None:
            start = Times(
                source=start,
                tzname=tzname)

        if stop is not None:
            stop = Times(
                source=stop,
                tzname=tzname)


        if (start and stop
                and stop <= start):
            stop = stop.shift('+1d')


        if start is not None:
            data['start'] = start.simple

        if stop is not None:
            data['stop'] = stop.simple


        super().__init__(**data)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .bridge import PhueBridge
from .device import PhueDevice
from .params import PhueDeviceParams
from .params import WhatPhueButtonParams
from .params import WhatPhueMotionParams
from .params import WhenPhueChangeParams
from .params import WhenPhueSceneParams
from .what import chck_phue_button
from .what import chck_phue_motion
from .what import what_phue_button
from .what import what_phue_motion
from .when import chck_phue_change
from .when import chck_phue_scene
from .when import when_phue_change
from .when import when_phue_scene



__all__ = [
    'PhueBridge',
    'PhueDevice',
    'PhueDeviceParams',
    'WhatPhueButtonParams',
    'WhatPhueMotionParams',
    'WhenPhueChangeParams',
    'WhenPhueSceneParams',
    'chck_phue_change',
    'when_phue_change',
    'chck_phue_scene',
    'when_phue_scene',
    'chck_phue_button',
    'what_phue_button',
    'chck_phue_motion',
    'what_phue_motion']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path



SAMPLES = (
    Path(__file__).parent
    .joinpath('samples'))



SCENE_PHIDS = [

    ('5808a516-aab3-3ec3'
     '-8eee-4db5152b07b5'),

    ('9678ff8b-d452-49f3'
     '-861c-74e5c5b2ca7c')]

SCENE_PATHS = [

    ('https://192.168.1.10'
     '/clip/v2/resource/scene'
     f'/{SCENE_PHIDS[0]}'),

    ('https://192.168.2.10'
     '/clip/v2/resource/scene'
     f'/{SCENE_PHIDS[1]}')]



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path
from typing import TYPE_CHECKING

from encommon.utils import save_text

from respx import MockRouter

from ..params import WhenPhueChangeParams
from ..params import WhenPhueSceneParams
from ...conftest import config_factory
from ...conftest import homie_factory
from ...homie import HomieWhen
from ...homie import HomieWhenParams

if TYPE_CHECKING:
    from ...homie import Homie



def test_when_phue_change(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """


    phue_change = (
        WhenPhueChangeParams(
            devices=['jupiter_motion'],
            sensors=['motion1']))

    params = HomieWhenParams(
        phue_change=phue_change)

    when = HomieWhen(homie, params)

    assert when.outcome is False


    phue_change = (
        WhenPhueChangeParams(
            devices=['jupiter_motion'],
            sensors=['motion1'],
            since=86400))

    params = HomieWhenParams(
        phue_change=phue_change)

    when = HomieWhen(homie, params)

    assert when.outcome is True



def test_when_phue_change_cover(
    tmp_path: Path,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param respx_mock: Object for mocking request operation.
    """

    config = (
        'phue_devices:\n'
        '  noexst:\n'
        '    bridge: jupiter\n'
        '    phid: noexst\n')


    Path.mkdir(
        tmp_path.joinpath('config'))

    save_text(
        f'{tmp_path}/config/test.yml',
        content=config)

    homie = homie_factory(
        config_factory(tmp_path),
        respx_mock)


    phue_change = (
        WhenPhueChangeParams(
            devices=['noexst'],
            sensors=['button1']))

    params = HomieWhenParams(
        phue_change=phue_change)

    when = HomieWhen(homie, params)

    assert when.outcome is False


    phue_change = (
        WhenPhueChangeParams(
            devices=['jupiter_switch'],
            sensors=['button1']))

    params = HomieWhenParams(
        phue_change=phue_change)

    when = HomieWhen(homie, params)

    assert when.outcome is False


    # Motion does not have buttons
    # this helps with the coverage
    phue_change = (
        WhenPhueChangeParams(
            devices=['jupiter_motion'],
            sensors=['button1']))

    params = HomieWhenParams(
        phue_change=phue_change)

    when = HomieWhen(homie, params)

    assert when.outcome is False


    phue_change = (
        WhenPhueChangeParams(
            devices=['jupiter_button'],
            sensors=['button4']))

    params = HomieWhenParams(
        phue_change=phue_change)

    when = HomieWhen(homie, params)

    assert when.outcome is False


    # Plugs do not have sensor but
    # this helps with the coverage
    phue_change = (
        WhenPhueChangeParams(
            devices=['jupiter_plug'],
            sensors=['motion1']))

    params = HomieWhenParams(
        phue_change=phue_change)

    when = HomieWhen(homie, params)

    assert when.outcome is False



def test_when_phue_scene(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """


    phue_scene = (
        WhenPhueSceneParams(
            group='jupiter_zone',
            scenes=['sleep']))

    params = HomieWhenParams(
        phue_scene=phue_scene)

    when = HomieWhen(homie, params)

    assert when.outcome is True


    phue_scene = (
        WhenPhueSceneParams(
            group='jupiter_room',
            scenes=['sleep']))

    params = HomieWhenParams(
        phue_scene=phue_scene)

    when = HomieWhen(homie, params)

    assert when.outcome is False



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import Optional
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .homie import Homie
    from .group import HomieGroup
    from .params import HomieSceneParams
    from ..philipshue.bridge import (
        _FETCH as PHUE_FETCH)



class HomieScene:
    """
    Normalize the scene parameter across multiple products.

    :param homie: Primary class instance for Homie Automate.
    :param name: Name of the object within the Homie config.
    """

    __homie: 'Homie'
    __params: 'HomieSceneParams'

    __name: str


    def __init__(
        self,
        homie: 'Homie',
        name: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        homie.log_d(
            base='HomieScene',
            name=name,
            status='initial')


        scenes = (
            homie.params.scenes)

        assert scenes is not None

        params = scenes[name]


        self.__homie = homie
        self.__params = params
        self.__name = name


        homie.log_d(
            base='HomieScene',
            name=name,
            status='created')


    @property
    def homie(
        self,
    ) -> 'Homie':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.__homie


    @property
    def params(
        self,
    ) -> 'HomieSceneParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def name(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__name


    def phue_source(
        self,
        group: 'HomieGroup',
    ) -> Optional['PHUE_FETCH']:
        """
        Return the dictionary containing the source from bridge.

        .. note::
           Scenes only exist within the groups on the bridge.

        :param group: Group from wherein the scene is located.
        :returns: Dictionary containing the source from bridge.
        """

        bridge = group.phue_bridge

        label = self.params.phue_label
        group_phid = group.phue_unique

        assert group_phid is not None

        return bridge.get_source(
            label=label,
            type='scene',
            grid=group_phid)


    def phue_unique(
        self,
        group: 'HomieGroup',
    ) -> Optional[str]:
        """
        Return the unique identifier of scene within the bridge.

        .. note::
           Scenes only exist within the groups on the bridge.

        :param group: Group from wherein the scene is located.
        :returns: Unique identifier of scene within the bridge.
        """

        source = (
            self.phue_source(group))

        if source is None:
            return None

        phid = source['id']

        assert isinstance(phid, str)

        return phid


    def phue_actual(
        self,
        group: 'HomieGroup',
    ) -> Optional[str]:
        """
        Return the actual name for the scene within the bridge.

        .. note::
           Scenes only exist within the groups on the bridge.

        :param group: Group from wherein the scene is located.
        :returns: Actual name for the scene within the bridge.
        """

        source = (
            self.phue_source(group))

        if source is None:
            return None

        metadata = source['metadata']
        name = metadata['name']

        assert isinstance(name, str)

        return name


    def phue_active(
        self,
        group: 'HomieGroup',
    ) -> Optional[bool]:
        """
        Return the boolean indicating when this is active scene.

        .. note::
           Scenes only exist within the groups on the bridge.

        :param group: Group from wherein the scene is located.
        :returns: Boolean indicating whether the scene active.
        """

        bridge = group.phue_bridge

        group_phid = group.phue_unique

        assert group_phid is not None

        crrnt_phid = (
            bridge.scene_get(group_phid))

        scene_phid = (
            self.phue_unique(group))

        return crrnt_phid == scene_phid


    def scene_set(
        self,
        group: 'HomieGroup',
    ) -> None:
        """
        Update the current group to activate the provided scene.

        :param group: Group from wherein the scene is located.
        """

        bridge = group.phue_bridge

        scene_phid = (
            self.phue_unique(group))

        assert scene_phid is not None

        bridge.scene_set(scene_phid)


    def homie_dumper(
        self,
        group: Optional['HomieGroup'] = None,
    ) -> dict[str, Any]:
        """
        Return the content related to the project dumper script.

        :param group: Group from wherein the scene is located.
        :returns: Content related to the project dumper script.
        """

        params = (
            self.params.model_dump())

        phue_actual: Optional[str] = None
        phue_unique: Optional[str] = None

        present: bool = False


        if group is not None:

            phue_actual = (
                self.phue_actual(group))

            phue_unique = (
                self.phue_unique(group))

            if phue_unique is not None:
                present = True


        return {
            'name': self.name,
            'present': present,
            'phue_unique': phue_unique,
            'phue_actual': phue_actual,
            'params': params}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from ..builtins import chck_time_period
from ..builtins import when_time_period
from ..philipshue import chck_phue_change
from ..philipshue import chck_phue_scene
from ..philipshue import when_phue_change
from ..philipshue import when_phue_scene
from ..ubiquiti import chck_ubiq_client
from ..ubiquiti import when_ubiq_client

if TYPE_CHECKING:
    from .homie import Homie
    from .params import HomieWhenParams



class HomieWhen:
    """
    Determine when condition is true and provide information.

    :param homie: Primary class instance for Homie Automate.
    :param params: Parameters for the conditional operations.
    """

    __homie: 'Homie'
    __params: 'HomieWhenParams'


    def __init__(
        self,
        homie: 'Homie',
        params: 'HomieWhenParams',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__homie = homie
        self.__params = params

        self.__validate_params()


    def __validate_params(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        if self.params.time_period:
            chck_time_period(self)

        if self.params.phue_change:
            chck_phue_change(self)

        if self.params.phue_scene:
            chck_phue_scene(self)

        if self.params.ubiq_client:
            chck_ubiq_client(self)


    @property
    def homie(
        self,
    ) -> 'Homie':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.__homie


    @property
    def params(
        self,
    ) -> 'HomieWhenParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def negate(
        self,
    ) -> bool:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.negate


    @property
    def family(
        self,
    ) -> str:
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.family


    @property
    def outcome(
        self,
    ) -> bool:
        """
        Return the boolean indicating whether condition matched.

        :returns: Boolean indicating whether condition matched.
        """

        matched: list[bool] = []

        if self.params.time_period:
            matched.append(
                when_time_period(self))

        if self.params.phue_change:
            matched.append(
                when_phue_change(self))

        if self.params.phue_scene:
            matched.append(
                when_phue_scene(self))

        if self.params.ubiq_client:
            matched.append(
                when_ubiq_client(self))

        outcome = bool(
            matched and all(matched))

        return (
            not outcome
            if self.negate is True
            else outcome)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import Timers
from encommon.types import sort_dict

from .action import HomieAction
from .desire import HomieDesire
from .group import HomieGroup
from .scene import HomieScene
from ..philipshue import PhueBridge
from ..philipshue import PhueDevice
from ..ubiquiti import UbiqClient
from ..ubiquiti import UbiqRouter

if TYPE_CHECKING:
    from ..config import Config
    from ..config import Params



_DESIRES = dict[str, list[HomieDesire]]
_DESIRED = dict[str, HomieDesire]

_ASPIRES = dict[str, list[HomieAction]]
_ASPIRED = dict[str, HomieAction]



class Homie:
    """
    Interact with supported devices to ensure desired state.

    :param config: Primary class instance for configuration.
    """

    __config: 'Config'
    __timers: Timers

    __phue_bridges: dict[str, PhueBridge]
    __phue_devices: dict[str, PhueDevice]
    __ubiq_routers: dict[str, UbiqRouter]
    __ubiq_clients: dict[str, UbiqClient]

    __groups: dict[str, HomieGroup]
    __scenes: dict[str, HomieScene]
    __desires: dict[str, HomieDesire]
    __actions: dict[str, HomieAction]


    def __init__(
        self,
        config: 'Config',
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        self.__config = config

        self.__make_timers()

        self.__phue_bridges = {}
        self.__phue_devices = {}
        self.__ubiq_routers = {}
        self.__ubiq_clients = {}

        self.__groups = {}
        self.__scenes = {}
        self.__desires = {}
        self.__actions = {}

        self.log_d(
            base='Homie',
            status='initial')


        self.__make_phue_bridges()
        self.__make_phue_devices()
        self.__make_ubiq_routers()
        self.__make_ubiq_clients()

        self.__make_groups()
        self.__make_scenes()
        self.__make_desires()
        self.__make_actions()


        self.__validate_params()

        self.log_i(
            base='Homie',
            status='created')


    def __validate_params(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        groups = self.groups
        phue_bridges = self.phue_bridges


        def _validate_groups() -> None:

            assert phue_bridges is not None

            for group in groups.values():

                params = group.params

                name = params.phue_bridge

                assert name in phue_bridges


        if groups is not None:
            _validate_groups()


    def __make_timers(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        cache = self.params.cache

        timers = Timers(store=cache)

        self.__timers = timers


    def __make_groups(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        params = self.params
        groups = params.groups

        if groups is None:
            return

        for name in groups.keys():

            group = HomieGroup(self, name)

            self.__groups |= {
                group.name: group}


    def __make_scenes(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        params = self.params
        scenes = params.scenes

        if scenes is None:
            return

        for name in scenes.keys():

            scene = HomieScene(self, name)

            self.__scenes |= {
                scene.name: scene}


    def __make_desires(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        params = self.params
        desires = params.desires

        if desires is None:
            return

        for name in desires.keys():

            desire = HomieDesire(self, name)

            self.__desires |= {
                desire.name: desire}


    def __make_actions(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        params = self.params
        actions = params.actions

        if actions is None:
            return

        for name in actions.keys():

            action = HomieAction(self, name)

            self.__actions |= {
                action.name: action}


    def __make_phue_bridges(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        params = self.params
        bridges = params.phue_bridges

        if bridges is None:
            return

        for name in bridges.keys():

            bridge = PhueBridge(self, name)

            self.__phue_bridges |= {
                bridge.name: bridge}


    def __make_phue_devices(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        params = self.params
        devices = params.phue_devices

        if devices is None:
            return

        for name in devices.keys():

            device = PhueDevice(self, name)

            self.__phue_devices |= {
                device.name: device}


    def __make_ubiq_routers(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        params = self.params
        routers = params.ubiq_routers

        if routers is None:
            return

        for name in routers.keys():

            router = UbiqRouter(self, name)

            self.__ubiq_routers |= {
                router.name: router}


    def __make_ubiq_clients(
        self,
    ) -> None:
        """
        Construct instances using the configuration parameters.
        """

        params = self.params
        clients = params.ubiq_clients

        if clients is None:
            return

        for name in clients.keys():

            client = UbiqClient(self, name)

            self.__ubiq_clients |= {
                client.name: client}


    def log(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        .. note::
           Uses method :py:meth:`encommon.config.Logger.log`.

        :param args: Positional arguments passed for downstream.
        :param kwargs: Keyword arguments for populating message.
        """

        config = self.config
        logger = config.logger

        logger.log(*args, **kwargs)


    def log_c(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        .. note::
           Uses method :py:meth:`encommon.config.Logger.log_c`.

        :param kwargs: Keyword arguments for populating message.
        """

        config = self.config
        logger = config.logger

        logger.log_c(**kwargs)


    def log_d(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        .. note::
           Uses method :py:meth:`encommon.config.Logger.log_d`.

        :param kwargs: Keyword arguments for populating message.
        """

        config = self.config
        logger = config.logger

        logger.log_d(**kwargs)


    def log_e(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        .. note::
           Uses method :py:meth:`encommon.config.Logger.log_e`.

        :param kwargs: Keyword arguments for populating message.
        """

        config = self.config
        logger = config.logger

        logger.log_e(**kwargs)


    def log_i(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        .. note::
           Uses method :py:meth:`encommon.config.Logger.log_i`.

        :param kwargs: Keyword arguments for populating message.
        """

        config = self.config
        logger = config.logger

        logger.log_i(**kwargs)


    def log_w(
        self,
        **kwargs: Any,
    ) -> None:
        """
        Pass the provided keyword arguments into logger object.

        .. note::
           Uses method :py:meth:`encommon.config.Logger.log_w`.

        :param kwargs: Keyword arguments for populating message.
        """

        config = self.config
        logger = config.logger

        logger.log_w(**kwargs)


    @property
    def config(
        self,
    ) -> 'Config':
        """
        Return the Config instance containing the configuration.

        :returns: Config instance containing the configuration.
        """

        return self.__config


    @property
    def params(
        self,
    ) -> 'Params':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.config.params


    @property
    def timers(
        self,
    ) -> Timers:
        """
        Return the timers instances defined within this instance.

        :returns: Timers instances defined within this instance.
        """

        return self.__timers


    @property
    def groups(
        self,
    ) -> dict[str, HomieGroup]:
        """
        Return the group instances defined within this instance.

        :returns: Group instances defined within this instance.
        """

        return dict(self.__groups)


    @property
    def rooms(
        self,
    ) -> dict[str, HomieGroup]:
        """
        Return the room instances defined within this instance.

        :returns: Room instances defined within this instance.
        """

        return {
            k: v for k, v in
            self.groups.items()
            if v.type == 'room'}


    @property
    def zones(
        self,
    ) -> dict[str, HomieGroup]:
        """
        Return the zone instances defined within this instance.

        :returns: Zone instances defined within this instance.
        """

        return {
            k: v for k, v in
            self.groups.items()
            if v.type == 'zone'}


    @property
    def scenes(
        self,
    ) -> dict[str, HomieScene]:
        """
        Return the scene instances defined within this instance.

        :returns: Scene instances defined within this instance.
        """

        return dict(self.__scenes)


    @property
    def desires(
        self,
    ) -> dict[str, HomieDesire]:
        """
        Return the desire instances defined within this instance.

        :returns: Desire instances defined within this instance.
        """

        return dict(self.__desires)


    @property
    def actions(
        self,
    ) -> dict[str, HomieAction]:
        """
        Return the action instances defined within this instance.

        :returns: Action instances defined within this instance.
        """

        return dict(self.__actions)


    @property
    def phue_bridges(
        self,
    ) -> dict[str, PhueBridge]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return dict(self.__phue_bridges)


    @property
    def phue_devices(
        self,
    ) -> dict[str, PhueDevice]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return dict(self.__phue_devices)


    @property
    def ubiq_routers(
        self,
    ) -> dict[str, UbiqRouter]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return dict(self.__ubiq_routers)


    @property
    def ubiq_clients(
        self,
    ) -> dict[str, UbiqClient]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return dict(self.__ubiq_clients)


    @property
    def desired(
        self,
    ) -> _DESIRED:
        """
        Return the related desired state for the desired groups.

        :returns: Related desired state for the desired groups.
        """

        params = self.params
        dryrun = params.dryrun

        desires: _DESIRES = {}


        def _append_desire() -> None:

            if _name not in desires:
                desires[_name] = []

            target = desires[_name]

            target.append(desire)


        items1 = self.desires.items()

        for name, desire in items1:

            if desire.outcome is False:

                if dryrun is False:
                    desire.delete_timer()

                continue

            if desire.delayed is True:
                continue

            groups = desire.groups

            for _name in groups:
                _append_desire()


        desired: _DESIRED = {}


        items2 = desires.items()

        for key, value in items2:

            value = sorted(
                value,
                key=lambda x: x.weight,
                reverse=True)

            desired[key] = value[0]


        return sort_dict(desired)


    def aspired(
        self,
        event: dict[str, Any],
    ) -> _ASPIRED:
        """
        Return the related actions matching the provided event.

        :param event: Event which was yielded from the stream.
        :returns: Related actions matching the provided event.
        """

        aspires: _ASPIRES = {}


        def _append_aspire() -> None:

            if _name not in aspires:
                aspires[_name] = []

            target = aspires[_name]

            target.append(action)


        items1 = self.actions.items()

        for name, action in items1:

            if action.outcome is False:
                continue

            matched = action.match(event)

            if matched is False:
                continue

            if action.paused is True:
                continue

            groups = action.groups

            for _name in groups:
                _append_aspire()


        aspired: _ASPIRED = {}


        items2 = aspires.items()

        for key, value in items2:

            value = sorted(
                value,
                key=lambda x: x.weight,
                reverse=True)

            aspired[key] = value[0]


        return sort_dict(aspired)


    def scene_get(
        self,
        group: HomieGroup,
    ) -> Optional[HomieScene]:
        """
        Return the current active scene when there is one active.

        :param group: Group from wherein the scene is located.
        :returns: Current active scene when there is one active.
        """

        scenes = self.scenes

        for scene in scenes.values():

            if scene.phue_active(group):
                return scene

        return None


    def scene_set(
        self,
        group: HomieGroup,
        scene: HomieScene,
    ) -> None:
        """
        Update the provided group to activate the provided scene.

        :param group: Group from wherein the scene is located.
        :param scene: Name of the Homie scene which is desired.
        """

        self.log_i(
            base='Homie',
            action='scene_set',
            group=group.name,
            scene=scene.name,
            status='attempt')

        group.scene_set(scene)

        self.log_i(
            base='Homie',
            action='scene_set',
            group=group.name,
            scene=scene.name,
            status='success')


    def homie_dumper(
        self,
    ) -> dict[str, Any]:
        """
        Return the content related to the project dumper script.

        :returns: Content related to the project dumper script.
        """

        dumped: dict[str, Any] = {
            'groups': {
                'rooms': sorted(self.rooms),
                'zones': sorted(self.zones)},
            'scenes': sorted(self.scenes),
            'desires': sorted(self.desires),
            'actions': sorted(self.actions)}

        _dumped: dict[str, Any]


        phue_bridges = (
            self.phue_bridges.values())

        _dumped = {}

        for bridge in phue_bridges:
            _dumped[bridge.name] = {
                'connect': bridge.connect}

        dumped['phue_bridges'] = _dumped


        phue_devices = (
            self.phue_devices.values())

        _dumped = {}

        for device in phue_devices:
            device.refresh_source()
            _dumped[device.name] = {
                'present': device.present,
                'connect': device.connect}

        dumped['phue_devices'] = _dumped


        ubiq_routers = (
            self.ubiq_routers.values())

        _dumped = {}

        for router in ubiq_routers:
            _dumped[router.name] = {
                'connect': router.connect}

        dumped['ubiq_routers'] = _dumped


        ubiq_clients = (
            self.ubiq_clients.values())

        _dumped = {}

        for client in ubiq_clients:
            client.refresh_source()
            _dumped[client.name] = {
                'present': client.present,
                'connect': client.connect}

        dumped['ubiq_clients'] = _dumped


        return dumped



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import loads
from typing import TYPE_CHECKING

from encommon import ENPYRWS
from encommon.types import inrepr
from encommon.types import instr
from encommon.utils import load_sample
from encommon.utils import prep_sample
from encommon.utils import read_text

from . import SAMPLES
from ...conftest import REPLACES
from ...philipshue.test import (
    SAMPLES as PHUE_SAMPLES)

if TYPE_CHECKING:
    from ..homie import Homie



def test_HomieAction(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    actions = homie.actions
    action = actions['jupiter']

    _events = loads(
        read_text(
            f'{PHUE_SAMPLES}/events'
            '/source.json'))


    attrs = list(action.__dict__)

    assert attrs == [
        '_HomieAction__homie',
        '_HomieAction__params',
        '_HomieAction__name',
        '_HomieAction__what',
        '_HomieAction__when']


    assert inrepr(
        'action.HomieAction object',
        action)

    assert hash(action) > 0

    assert instr(
        'action.HomieAction object',
        action)


    assert action.homie is homie

    assert action.params is not None

    assert action.name == 'jupiter'

    assert len(action.whats) == 2

    assert len(action.whens) == 1

    assert len(action.groups) == 2

    assert action.scene == 'awake'

    assert action.weight == 10

    assert action.what is not None

    assert action.when is not None

    assert len(action.when) == 1

    assert action.paused is False

    assert action.outcomes == {
        'default': [True]}

    assert action.outcome is True

    matches = action.matches(_events[0])

    assert matches == [True, False]

    match = action.match(_events[0])

    assert match is True


    sample_path = (
        f'{SAMPLES}/action'
        '/dumper.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=action.homie_dumper(),
        replace=REPLACES)

    expect = prep_sample(
        content=action.homie_dumper(),
        replace=REPLACES)

    assert sample == expect



def test_HomieAction_cover(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    actions = homie.actions
    jupiter = actions['jupiter']
    default = actions['default']

    _events = loads(
        read_text(
            f'{PHUE_SAMPLES}/events'
            '/source.json'))


    aspired = homie.aspired(_events[0])

    assert aspired == {
        'jupiter_room': jupiter,
        'jupiter_zone': jupiter,
        'neptune_room': default,
        'neptune_zone': default}


    jupiter.update_timer()

    aspired = homie.aspired(_events[0])

    assert aspired == {
        'jupiter_room': default,
        'jupiter_zone': default,
        'neptune_room': default,
        'neptune_zone': default}


    jupiter.delete_timer()

    aspired = homie.aspired(_events[0])

    assert aspired == {
        'jupiter_room': jupiter,
        'jupiter_zone': jupiter,
        'neptune_room': default,
        'neptune_zone': default}


    aspired = homie.aspired(_events[1])

    assert aspired == {
        'jupiter_room': jupiter,
        'jupiter_zone': jupiter}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import Times

if TYPE_CHECKING:
    from ..homie import HomieWhen



def when_time_period(
    when: 'HomieWhen',
) -> bool:
    """
    Return the boolean indicating whether condition matched.

    :param when: Primary class instance for the conditonal.
    :returns: Boolean indicating whether condition matched.
    """

    params = (
        when.params.time_period)

    assert params is not None

    start = params.start
    stop = params.stop
    tzname = params.tzname

    anchor = Times(tzname=tzname)

    assert (
        start is not None
        or stop is not None)


    _start: Optional[Times] = (
        Times(start)
        if start is not None
        else None)

    _stop: Optional[Times] = (
        Times(stop)
        if stop is not None
        else None)


    matched = True

    if (_start is not None
            and anchor < _start):
        matched = False

    if (_stop is not None
            and anchor > _stop):
        matched = False


    return matched



def chck_time_period(
    when: 'HomieWhen',
) -> None:
    """
    Return the boolean indicating whether conditional valid.

    :param when: Primary class instance for the conditonal.
    """

    params = (
        when.params.time_period)

    assert params is not None

    start = params.start
    stop = params.stop
    tzname = params.tzname


    if start is not None:
        Times(start, tzname=tzname)

    if stop is not None:
        Times(stop, tzname=tzname)



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon import ENPYRWS
from encommon.types import inrepr
from encommon.types import instr
from encommon.utils import load_sample
from encommon.utils import prep_sample

from . import SAMPLES
from ...conftest import REPLACES

if TYPE_CHECKING:
    from ..homie import Homie



def test_HomieGroup(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    groups = homie.groups
    group = groups['jupiter_room']


    attrs = list(group.__dict__)

    assert attrs == [
        '_HomieGroup__homie',
        '_HomieGroup__params',
        '_HomieGroup__name']


    assert inrepr(
        'group.HomieGroup object',
        group)

    assert hash(group) > 0

    assert instr(
        'group.HomieGroup object',
        group)


    assert group.homie is homie

    assert group.params is not None

    assert group.name == 'jupiter_room'

    assert group.type == 'room'

    assert group.phue_bridge is not None

    assert group.phue_source is not None

    assert group.phue_unique == (
        'da9031f1-6229-3e2d'
        '-b143-64b2c8223915')

    assert group.phue_actual == 'Jupiter'


    sample_path = (
        f'{SAMPLES}/group/dumper.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=group.homie_dumper(),
        replace=REPLACES)

    expect = prep_sample(
        content=group.homie_dumper(),
        replace=REPLACES)

    assert sample == expect



def test_HomieGroup_cover(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    groups = homie.groups
    group = groups['jupiter_room']
    params = group.params

    params.phue_label = 'noexst'

    assert not group.phue_actual
    assert not group.phue_unique



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .action import HomieAction
from .desire import HomieDesire
from .group import HomieGroup
from .homie import Homie
from .params import HomieActionParams
from .params import HomieDesireParams
from .params import HomieGroupParams
from .params import HomieSceneParams
from .params import HomieWhatParams
from .params import HomieWhenParams
from .scene import HomieScene
from .what import HomieWhat
from .when import HomieWhen



__all__ = [
    'Homie',
    'HomieAction',
    'HomieActionParams',
    'HomieDesire',
    'HomieDesireParams',
    'HomieGroup',
    'HomieGroupParams',
    'HomieScene',
    'HomieSceneParams',
    'HomieWhen',
    'HomieWhenParams',
    'HomieWhat',
    'HomieWhatParams']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from .client import UbiqClient
from .params import UbiqClientParams
from .params import WhenUbiqClientParams
from .router import UbiqRouter
from .when import chck_ubiq_client
from .when import when_ubiq_client



__all__ = [
    'UbiqClient',
    'UbiqClientParams',
    'UbiqRouter',
    'WhenUbiqClientParams',
    'chck_ubiq_client',
    'when_ubiq_client']



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from contextlib import suppress
from typing import Any
from typing import Optional

from encommon.config import Config as _Config
from encommon.types import merge_dicts
from encommon.utils.common import PATHABLE

from .params import Params



class Config(_Config):
    """
    Contain the configurations from the arguments and files.

    :param files: Complete or relative path to config files.
    :param cargs: Configuration arguments in dictionary form,
        which will override contents from the config files.
    :param sargs: Additional arguments on the command line.
    """

    def __init__(
        self,
        files: Optional[PATHABLE] = None,
        cargs: Optional[dict[str, Any]] = None,
        sargs: Optional[dict[str, Any]] = None,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        super().__init__(
            files=files,
            cargs=cargs,
            sargs=sargs,
            model=Params)


    @property
    def params(
        self,
    ) -> Params:
        """
        Return the Pydantic model containing the configuration.

        .. warning::
           This method completely overrides the parent but is
           based on that code, would be unfortunate if upstream
           changes meant this breaks or breaks something else.

        :returns: Pydantic model containing the configuration.
        """

        params = self.__params

        if isinstance(params, Params):
            return params


        merged = self.files.merged

        merge_dicts(
            dict1=merged,
            dict2=self.cargs,
            force=True)


        update = False

        with suppress(AttributeError):

            _merged = self.paths.merged

            values = _merged.values()

            for merge in values:
                merge_dicts(merged, merge)

            update = True


        params = self.model(**merged)

        assert isinstance(params, Params)


        if update is True:
            self.__params = params

        return params



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..homie import HomieWhen



def when_ubiq_client(
    when: 'HomieWhen',
) -> bool:
    """
    Return the boolean indicating whether condition matched.

    :param when: Primary class instance for the conditonal.
    :returns: Boolean indicating whether condition matched.
    """

    params = (
        when.params.ubiq_client)

    assert params is not None

    _routers = params.routers
    _clients = params.clients
    since = params.since

    routers = when.homie.ubiq_routers
    clients = when.homie.ubiq_clients


    def _append_outcome() -> None:

        _router_names = (
            _routers or routers)

        assert connect is not False
        assert lseen is not None

        for name in _router_names:

            _lseen = lseen.get(name)

            if (_lseen is not None
                    and _lseen.since < since):
                outcomes.append(True)

            elif name in connect:
                outcomes.append(True)


    outcomes: list[bool] = []

    for name in _clients:

        client = clients[name]
        client.refresh_source()

        present = client.present
        connect = client.connect
        lseen = client.lastseen

        if present is False:
            outcomes.append(False)
            continue

        if connect is False:
            outcomes.append(False)
            continue

        _append_outcome()


    return any(outcomes)



def chck_ubiq_client(
    when: 'HomieWhen',
) -> None:
    """
    Return the boolean indicating whether conditional valid.

    :param when: Primary class instance for the conditonal.
    """

    params = (
        when.params.ubiq_client)

    assert params is not None

    routers = (
        when.homie.ubiq_routers)


    _routers = (
        params.routers or [])

    for name in _routers:
        assert name in routers



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from typing import Any
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import Duration
from encommon.times import Times
from encommon.types import getate

if TYPE_CHECKING:
    from .bridge import _FETCH
    from .bridge import PhueBridge
    from .params import PhueDeviceParams
    from ..homie import Homie



_CHANGED = dict[str, Times | None] | Literal[False]
_SENSORS = Optional[dict[str, str]]



class PhueDevice:
    """
    Contain the relevant attributes about the related device.

    :param homie: Primary class instance for Homie Automate.
    :param name: Name of the object within the Homie config.
    """

    __homie: 'Homie'
    __params: 'PhueDeviceParams'

    __name: str

    __source: Optional['_FETCH']
    __bridge: Optional['PhueBridge']


    def __init__(
        self,
        homie: 'Homie',
        name: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        homie.log_d(
            base='PhueDevice',
            name=name,
            status='initial')


        devices = (
            homie.params
            .phue_devices)

        assert devices is not None

        params = devices[name]


        self.__homie = homie
        self.__params = params
        self.__name = name
        self.__source = None
        self.__bridge = None


        self.__validate_params()

        homie.log_d(
            base='PhueDevice',
            name=name,
            status='created')


    def __validate_params(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        bridges = self.homie.phue_bridges
        bridge_name = self.params.bridge

        assert bridges is not None
        assert bridge_name in bridges


    def refresh_source(
        self,
    ) -> None:
        """
        Update the class instance from cached upstream response.
        """

        homie = self.homie
        bridges = homie.phue_bridges
        params = self.params

        assert bridges is not None

        self.__source = None
        self.__bridge = None


        expect = params.bridge

        for bridge in bridges.values():

            name = bridge.name

            if expect != name:
                continue


            found = bridge.get_source(
                phid=params.phid,
                label=params.label,
                type='device')

            if found is None:
                continue


            _type = found['type']

            assert _type == 'device'

            assert not self.__source

            self.__source = found
            self.__bridge = bridge


    @property
    def homie(
        self,
    ) -> 'Homie':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.__homie


    @property
    def params(
        self,
    ) -> 'PhueDeviceParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def name(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__name


    @property
    def bridge(
        self,
    ) -> Optional['PhueBridge']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        if self.__bridge is None:
            return None

        return self.__bridge


    @property
    def source(
        self,
    ) -> Optional['_FETCH']:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        if self.__source is None:
            return None

        return deepcopy(self.__source)


    @property
    def unique(
        self,
    ) -> Optional[str]:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        source = self.__source

        if source is None:
            return None

        phid = source['id']

        assert isinstance(phid, str)

        return phid


    @property
    def present(
        self,
    ) -> bool:
        """
        Return the boolean indicating device present on bridge.

        :returns: Boolean indicating device present on bridge.
        """

        return self.__source is not None


    @property
    def connect(
        self,
    ) -> bool:
        """
        Return the boolean indicating device connected to bridge.

        :returns: Boolean indicating device connected to bridge.
        """

        source = self.__source

        key = 'zigbee_connectivity'

        if source is None:
            return False


        services = source['services']

        connect = False

        for service in services:

            rtype = service['rtype']

            if rtype != key:
                continue

            origin = service['_source']
            status = origin['status']

            if status == 'connected':
                connect = True


        return connect


    @property
    def changed(
        self,
    ) -> _CHANGED:
        """
        Return the related information for services have changed.

        :returns: Related information for services have changed.
        """

        keys = ['button', 'motion']

        source = self.__source

        if source is None:
            return False


        services = source['services']

        changed: dict[str, Times | None] = {}
        indices = {x: 0 for x in keys}

        for service in services:

            _rtype = service['rtype']

            if _rtype not in keys:
                continue

            indices[_rtype] += 1

            _changed: Optional[str] = None

            _source = service['_source']

            if _rtype == 'motion':
                _changed = getate(
                    _source['motion'],
                    'motion_report/changed')

            elif _rtype == 'button':
                _changed = getate(
                    _source['button'],
                    'button_report/updated')

            index = indices[_rtype]

            key = f'{_rtype}{index}'

            changed[key] = (
                Times(_changed)
                if _changed is not None
                else None)


        values = changed.values()

        return changed if any(values) else False


    @property
    def sensors(
        self,
    ) -> _SENSORS:
        """
        Return the mapping for sensor name to unique identifier.

        :returns: Mapping for sensor name to unique identifier.
        """

        keys = ['button', 'motion']

        source = self.__source

        if source is None:
            return None


        services = source['services']

        sensors: dict[str, str] = {}
        indices = {x: 0 for x in keys}

        for service in services:

            _rid = service['rid']
            _rtype = service['rtype']

            if _rtype not in keys:
                continue

            indices[_rtype] += 1

            index = indices[_rtype]

            key = f'{_rtype}{index}'

            sensors[key] = _rid


        return sensors or None


    def homie_dumper(
        self,
    ) -> dict[str, Any]:
        """
        Return the content related to the project dumper script.

        :returns: Content related to the project dumper script.
        """

        params = (
            self.params.model_dump())

        self.refresh_source()

        changed: dict[str, Any] = {}


        if self.changed is not False:

            items = self.changed.items()

            for key, value in items:

                if value is None:
                    continue

                value = Times(value)

                durate = Duration(
                    value.since, groups=2)

                changed[key] = {
                    'stamp': value,
                    'since': durate}


        return {
            'name': self.name,
            'unique': self.unique,
            'present': self.present,
            'connect': self.connect,
            'changed': changed or False,
            'params': params}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..homie import HomieWhen



def when_phue_change(
    when: 'HomieWhen',
) -> bool:
    """
    Return the boolean indicating whether condition matched.

    :param when: Primary class instance for the conditonal.
    :returns: Boolean indicating whether condition matched.
    """

    params = (
        when.params.phue_change)

    assert params is not None

    _devices = params.devices
    _sensors = params.sensors
    since = params.since

    devices = (
        when.homie.phue_devices)


    def _append_outcome() -> None:

        assert changed is not False

        items = changed.items()

        for key, value in items:

            if key not in _sensors:
                continue

            if value is None:
                continue

            _since = value.since

            if _since > since:
                continue

            outcomes.append(True)


    outcomes: list[bool] = []

    for name in _devices:

        device = devices[name]
        device.refresh_source()

        present = device.present
        connect = device.connect
        changed = device.changed

        if not present:
            outcomes.append(False)
            continue

        if not connect:
            outcomes.append(False)
            continue

        if not changed:
            outcomes.append(False)
            continue

        _append_outcome()


    return any(outcomes)



def chck_phue_change(
    when: 'HomieWhen',
) -> None:
    """
    Return the boolean indicating whether conditional valid.

    :param when: Primary class instance for the conditonal.
    """

    params = (
        when.params.phue_change)

    assert params is not None

    devices = (
        when.homie.phue_devices)


    _devices = params.devices

    for name in _devices:
        assert name in devices



def when_phue_scene(
    when: 'HomieWhen',
) -> bool:
    """
    Return the boolean indicating whether condition matched.

    :param when: Primary class instance for the conditonal.
    :returns: Boolean indicating whether condition matched.
    """

    homie = when.homie

    params = (
        when.params.phue_scene)

    assert params is not None

    _group = params.group
    _scenes = params.scenes


    scene = homie.scene_get(
        homie.groups[_group])


    if scene is None:
        return False

    return scene.name in _scenes



def chck_phue_scene(
    when: 'HomieWhen',
) -> None:
    """
    Return the boolean indicating whether conditional valid.

    :param when: Primary class instance for the conditonal.
    """

    params = (
        when.params.phue_scene)

    assert params is not None

    groups = when.homie.groups
    scenes = when.homie.scenes


    _group = params.group

    assert _group in groups


    _scenes = params.scenes

    for name in _scenes:
        assert name in scenes



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from pathlib import Path

from _pytest.logging import LogCaptureFixture

from encommon import ENPYRWS
from encommon.types import inrepr
from encommon.types import instr
from encommon.utils import load_sample
from encommon.utils import prep_sample

from httpx import Response

from respx import MockRouter

from . import SAMPLES
from ..homie import Homie
from ...config import Config
from ...conftest import REPLACES
from ...philipshue.test import (
    SCENE_PATHS as PHUE_SCENE_PATHS)



def test_Homie(
    homie: Homie,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    desires = homie.desires
    desire = desires['default']


    attrs = list(homie.__dict__)

    assert attrs == [
        '_Homie__config',
        '_Homie__timers',
        '_Homie__phue_bridges',
        '_Homie__phue_devices',
        '_Homie__ubiq_routers',
        '_Homie__ubiq_clients',
        '_Homie__groups',
        '_Homie__scenes',
        '_Homie__desires',
        '_Homie__actions']


    assert inrepr(
        'homie.Homie object',
        homie)

    assert hash(homie) > 0

    assert instr(
        'homie.Homie object',
        homie)


    assert homie.config is not None

    assert homie.params is not None

    assert homie.timers is not None

    assert len(homie.groups) == 4

    assert len(homie.rooms) == 2

    assert len(homie.zones) == 2

    assert len(homie.scenes) == 5

    assert len(homie.desires) == 4

    assert len(homie.phue_bridges) == 2

    assert len(homie.phue_devices) == 10

    assert len(homie.ubiq_routers) == 2

    assert len(homie.ubiq_clients) == 6

    assert homie.desired == {
        'jupiter_room': desire,
        'jupiter_zone': desire,
        'neptune_room': desire,
        'neptune_zone': desire}


    sample_path = (
        f'{SAMPLES}/homie/dumper.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=homie.homie_dumper(),
        replace=REPLACES)

    expect = prep_sample(
        content=homie.homie_dumper(),
        replace=REPLACES)

    assert sample == expect



def test_Homie_logger(
    homie: Homie,
    caplog: LogCaptureFixture,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param caplog: pytest object for capturing log message.
    """

    logger = homie.config.logger


    logger.start()

    homie.log_d(message='pytest')
    homie.log_c(message='pytest')
    homie.log_e(message='pytest')
    homie.log_i(message='pytest')
    homie.log_w(message='pytest')

    homie.log(
        level='debug',
        message='custom')

    logger.stop()

    output = caplog.record_tuples

    assert len(output) == 6


    homie.log_d(message='pytest')
    homie.log_c(message='pytest')
    homie.log_e(message='pytest')
    homie.log_i(message='pytest')
    homie.log_w(message='pytest')

    output = caplog.record_tuples

    assert len(output) == 6



def test_Homie_scene(
    homie: Homie,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    :param respx_mock: Object for mocking request operation.
    """


    for path in PHUE_SCENE_PATHS:

        (respx_mock
         .put(path)
         .mock(Response(200)))


    scene = homie.scene_get(
        homie.groups['jupiter_zone'])

    assert scene is not None
    assert scene.name == 'sleep'

    homie.scene_set(
        homie.groups['jupiter_room'],
        homie.scenes['awake'])


    scene = homie.scene_get(
        homie.groups['neptune_zone'])

    assert scene is not None
    assert scene.name == 'sleep'

    homie.scene_set(
        homie.groups['neptune_room'],
        homie.scenes['awake'])


    scene = homie.scene_get(
        homie.groups['jupiter_room'])

    assert scene is None



def test_Homie_cover(
    tmp_path: Path,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    """

    Homie(Config())



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Literal
from typing import Optional

from pydantic import BaseModel

from ..builtins import WhenTimePeriodParams
from ..philipshue import WhatPhueButtonParams
from ..philipshue import WhatPhueMotionParams
from ..philipshue import WhenPhueChangeParams
from ..philipshue import WhenPhueSceneParams
from ..ubiquiti import WhenUbiqClientParams



GROUP_TYPES = Literal['room', 'zone']



class HomieGroupParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    .. note::
       Value is required for ``phue_label`` but may one day
       become optional, should another product be supported.

    :param type: What type of group does the group represent.
    :param phue_bridge: Which bridge will group be found on.
    :param phue_label: Value for matching the name of group.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    type: GROUP_TYPES

    phue_bridge: str
    phue_label: str



class HomieSceneParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    .. note::
       Value is required for ``phue_label`` but may one day
       become optional, should another product be supported.

    :param phue_label: Value for matching the name of scene.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    phue_label: str



class HomieWhenParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    .. warning::
       See documentation on the ``family`` input parameter.

    :param negate: Whether or not final outcome is inverted.
    :param family: Aggregate the conditionals where `any` in
        the family may be `True`. Note for `default` group,
        `all` of the conditionals within must match `True`.
    :param time_period: Pamaters for the conditional plugin.
    :param phue_change: Pamaters for the conditional plugin.
    :param phue_scene: Pamaters for the conditional plugin.
    :param ubiq_client: Pamaters for the conditional plugin.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    negate: bool = False
    family: str = 'default'

    time_period: Optional[WhenTimePeriodParams] = None
    phue_change: Optional[WhenPhueChangeParams] = None
    phue_scene: Optional[WhenPhueSceneParams] = None
    ubiq_client: Optional[WhenUbiqClientParams] = None



class HomieDesireParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    :param groups: Names of Homie groups which are in scope.
    :param scene: Name of the Homie scene which is desired.
    :param weight: Useful when other conditionals match one
        of the provided groups, will determine precedence.
    :param delay: Delay before taking action on scene change.
        This depends on the state file being defined to work
        between script executions, otherwise delay breaks.
    :param when: List of conditionals for determining match.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    groups: list[str]
    scene: str

    weight: int = 0
    delay: int = 0

    when: Optional[list[HomieWhenParams]] = None



class HomieWhatParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    :param phue_button: Parameters for use in action plugin.
    :param phue_motion: Parameters for use in action plugin.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    phue_button: Optional[WhatPhueButtonParams] = None
    phue_motion: Optional[WhatPhueMotionParams] = None



class HomieActionParams(BaseModel, extra='forbid'):
    """
    Process and validate the Homie configuration parameters.

    :param groups: Names of Homie groups which are in scope.
    :param scene: Name of the Homie scene which is desired.
    :param weight: Useful when other conditionals match one
        of the provided groups, will determine precedence.
    :param pause: Delay before performing the same action.
        This depends on the state file being defined to work
        between script executions, otherwise delay breaks.
    :param what: List of events which can trigger operation.
    :param when: List of conditionals for determining match.
    :param data: Keyword arguments passed to Pydantic model.
        Parameter is picked up by autodoc, please ignore.
    """

    groups: list[str]
    scene: str

    weight: int = 0
    pause: int = 5

    what: Optional[list[HomieWhatParams]] = None
    when: Optional[list[HomieWhenParams]] = None



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import TYPE_CHECKING

from encommon import ENPYRWS
from encommon.types import inrepr
from encommon.types import instr
from encommon.utils import load_sample
from encommon.utils import prep_sample

from . import SAMPLES
from ...conftest import REPLACES

if TYPE_CHECKING:
    from ..homie import Homie



def test_HomieScene(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    scenes = homie.scenes
    groups = homie.groups
    scene = scenes['awake']
    group = groups['jupiter_room']


    attrs = list(scene.__dict__)

    assert attrs == [
        '_HomieScene__homie',
        '_HomieScene__params',
        '_HomieScene__name']


    assert inrepr(
        'scene.HomieScene object',
        scene)

    assert hash(scene) > 0

    assert instr(
        'scene.HomieScene object',
        scene)


    assert scene.homie is homie

    assert scene.params is not None

    assert scene.name == 'awake'

    phue_source = (
        scene.phue_source(group))

    phue_unique = (
        scene.phue_unique(group))

    phue_actual = (
        scene.phue_actual(group))

    assert phue_source is not None

    assert phue_unique == (
        '5808a516-aab3-3ec3'
        '-8eee-4db5152b07b5')

    assert phue_actual == 'Awake'


    sample_path = (
        f'{SAMPLES}/scene/dumper.json')

    sample = load_sample(
        path=sample_path,
        update=ENPYRWS,
        content=scene.homie_dumper(group),
        replace=REPLACES)

    expect = prep_sample(
        content=scene.homie_dumper(group),
        replace=REPLACES)

    assert sample == expect



def test_HomieScene_cover(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    scenes = homie.scenes
    groups = homie.groups
    scene = scenes['awake']
    group = groups['jupiter_room']

    scene.params.phue_label = 'noexst'

    assert not (
        scene.phue_unique(group))

    assert not (
        scene.phue_actual(group))


    assert scene.homie_dumper() == {
        'name': 'awake',
        'present': False,
        'params': {
            'phue_label': 'noexst'},
        'phue_actual': None,
        'phue_unique': None}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from copy import deepcopy
from typing import Any
from typing import Literal
from typing import Optional
from typing import TYPE_CHECKING

from encommon.times import Timer
from encommon.times import Times
from encommon.types import merge_dicts
from encommon.types import striplower

from enconnect.ubiquiti import Router
from enconnect.ubiquiti import RouterParams

from urllib3 import disable_warnings
from urllib3.exceptions import InsecureRequestWarning

if TYPE_CHECKING:
    from ..homie import Homie



_FETCH = dict[str, Any]
_RAWDEV = dict[str, dict[str, Any]]



disable_warnings(
    category=InsecureRequestWarning)



class UbiqRouter:
    """
    Contain the relevant attributes about the related device.

    :param homie: Primary class instance for Homie Automate.
    :param name: Name of the object within the Homie config.
    """

    __homie: 'Homie'
    __params: RouterParams
    __router: Router

    __name: str

    __fetched: Optional[dict[str, _FETCH]]
    __timer: Timer
    __merged: Optional[_RAWDEV]


    def __init__(
        self,
        homie: 'Homie',
        name: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        homie.log_d(
            base='UbiqRouter',
            name=name,
            status='initial')


        routers = (
            homie.params
            .ubiq_routers)

        assert routers is not None

        params = routers[name]


        self.__homie = homie
        self.__params = params
        self.__router = Router(params)
        self.__name = name
        self.__fetched = None
        self.__merged = None


        self.__timer = Timer(
            60, start='-60s')


        homie.log_d(
            base='UbiqRouter',
            name=name,
            status='created')


    @property
    def homie(
        self,
    ) -> 'Homie':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.__homie


    @property
    def params(
        self,
    ) -> RouterParams:
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def router(
        self,
    ) -> Router:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__router


    @property
    def name(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__name


    @property
    def connect(
        self,
    ) -> bool:
        """
        Return the boolean indicating connection is established.

        :returns: Boolean indicating connection is established.
        """

        return bool(self.__merged)


    @property
    def fetched(
        self,
    ) -> _FETCH:
        """
        Collect the complete dump of all known clients in router.

        :returns: Complete dump of all known clients in router.
        """

        fetched = self.__fetched
        timer = self.__timer
        router = self.__router
        request = router.request_proxy

        ready = timer.ready(False)

        if fetched and not ready:
            return deepcopy(fetched)

        runtime = Times()


        response = request(
            'get', 'rest/user')

        historic = response.json()

        assert isinstance(historic, dict)


        response = request(
            'get', 'stat/sta')

        realtime = response.json()

        assert isinstance(realtime, dict)


        fetched = {
            'historic': historic,
            'realtime': realtime}


        self.homie.log_i(
            base='UbiqRouter',
            name=self.name,
            action='fetch',
            elapsed=runtime.since,
            status='success')


        self.__fetched = fetched
        self.__merged = None

        timer.update()

        return deepcopy(fetched)


    @property
    def merged(
        self,
    ) -> _RAWDEV:
        """
        Process the response and perform common transformations.

        :returns: Compiled response from the upstream endpoint.
        """

        staged: _RAWDEV = {}
        source: _RAWDEV = {}

        merged = self.__merged

        if merged is not None:
            return deepcopy(merged)

        fetched = self.fetched


        historic = (
            fetched['historic']['data'])

        realtime = (
            fetched['realtime']['data'])


        def _fixtimes(
            fetch: _FETCH,
        ) -> None:

            if 'first_seen' in item:
                item['first_seen'] = [
                    item['first_seen']]

            if 'last_seen' in item:
                item['last_seen'] = [
                    item['last_seen']]


        def _combine(
            target: Literal['historic', 'realtime'],
        ) -> None:

            ubid = item['_id']

            if ubid not in staged:
                staged[ubid] = {}

            source = staged[ubid]

            assert target not in source

            _fixtimes(item)

            source[target] = (
                deepcopy(item))


        for item in historic:
            _combine('historic')

        for item in realtime:
            _combine('realtime')


        items = staged.items()

        for ubid, fetch in items:

            historic = (
                fetch.get('historic', {}))

            realtime = (
                fetch.get('realtime', {}))

            _fetch: _FETCH = (
                deepcopy(historic))

            merge_dicts(
                dict1=_fetch,
                dict2=deepcopy(realtime),
                force=True)

            source[ubid] = _fetch | {
                '_historic': historic,
                '_realtime': _fetch}


        self.__merged = source

        return deepcopy(source)


    def get_source(
        self,
        ubid: Optional[str] = None,
        label: Optional[str] = None,
    ) -> Optional[_FETCH]:
        """
        Enumerate and collect information from cached response.

        :param ubid: Used for filtering resources for matching.
        :param label: Used for filtering resources for matching.
        :returns: Information for matching client from upstream.
        """

        assert ubid or label

        if ubid is not None:
            return self.get_source_ubid(ubid)

        if label is not None:
            return self.get_source_label(label)

        return None  # NOCVR


    def get_source_ubid(
        self,
        ubid: Optional[str] = None,
    ) -> Optional[_FETCH]:
        """
        Enumerate and collect information from cached response.

        :param ubid: Used for filtering resources for matching.
        :returns: Information for matching client from upstream.
        """

        found: list[_FETCH] = []

        items = self.merged.items()

        for _ubid, fetch in items:

            values = [striplower(_ubid)]

            if 'mac' in fetch:
                values.append(
                    striplower(fetch['mac']))

            if 'ip' in fetch:
                values.append(
                    striplower(fetch['ip']))

            if ubid not in values:
                continue

            found.append(fetch)

        assert len(found) in [0, 1]

        return found[0] if found else None


    def get_source_label(
        self,
        label: str,
    ) -> Optional[_FETCH]:
        """
        Enumerate and collect information from cached response.

        :param label: Used for filtering resources for matching.
        :returns: Information for matching client from upstream.
        """

        found: list[_FETCH] = []

        label = striplower(label)

        items = self.merged.items()

        for _ubid, fetch in items:

            values: list[str] = []

            if fetch.get('hostname'):
                values.append(
                    striplower(fetch['hostname']))

            if fetch.get('name'):
                values.append(
                    striplower(fetch['name']))

            if label not in values:
                continue

            found.append(fetch)

        assert len(found) in [0, 1]

        return found[0] if found else None


    def homie_dumper(
        self,
    ) -> dict[str, Any]:
        """
        Return the content related to the project dumper script.

        :returns: Content related to the project dumper script.
        """

        params = deepcopy(
            self.params.model_dump())

        params['password'] = (
            '*' * len(params['password']))

        return {
            'name': self.name,
            'connect': self.connect,
            'params': params}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from typing import Any
from typing import Optional
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .homie import Homie
    from .params import GROUP_TYPES
    from .params import HomieGroupParams
    from .scene import HomieScene
    from ..philipshue import PhueBridge
    from ..philipshue.bridge import (
        _FETCH as PHUE_FETCH)



class HomieGroup:
    """
    Normalize the group parameter across multiple products.

    :param homie: Primary class instance for Homie Automate.
    :param name: Name of the object within the Homie config.
    """

    __homie: 'Homie'
    __params: 'HomieGroupParams'

    __name: str


    def __init__(
        self,
        homie: 'Homie',
        name: str,
    ) -> None:
        """
        Initialize instance for class using provided parameters.
        """

        homie.log_d(
            base='HomieGroup',
            name=name,
            status='initial')


        groups = (
            homie.params.groups)

        assert groups is not None

        params = groups[name]


        self.__homie = homie
        self.__params = params
        self.__name = name


        self.__validate_params()

        homie.log_d(
            base='HomieGroup',
            name=name,
            status='created')


    def __validate_params(
        self,
    ) -> None:
        """
        Perform advanced validation on the parameters provided.
        """

        bridges = self.homie.phue_bridges
        bridge_name = self.params.phue_bridge

        assert bridges is not None
        assert bridge_name in bridges


    @property
    def homie(
        self,
    ) -> 'Homie':
        """
        Return the Homie instance to which this instance belongs.

        :returns: Homie instance to which this instance belongs.
        """

        return self.__homie


    @property
    def params(
        self,
    ) -> 'HomieGroupParams':
        """
        Return the Pydantic model containing the configuration.

        :returns: Pydantic model containing the configuration.
        """

        return self.__params


    @property
    def name(
        self,
    ) -> str:
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        return self.__name


    @property
    def type(
        self,
    ) -> 'GROUP_TYPES':
        """
        Return the value for the attribute from params instance.

        :returns: Value for the attribute from params instance.
        """

        return self.params.type


    @property
    def phue_bridge(
        self,
    ) -> 'PhueBridge':
        """
        Return the value for the attribute from class instance.

        :returns: Value for the attribute from class instance.
        """

        bridges = self.homie.phue_bridges
        bridge_name = self.params.phue_bridge

        assert bridges is not None

        return bridges[bridge_name]


    @property
    def phue_source(
        self,
    ) -> Optional['PHUE_FETCH']:
        """
        Return the dictionary containing the source from bridge.

        :returns: Dictionary containing the source from bridge.
        """

        bridge = self.phue_bridge

        return bridge.get_source(
            label=self.params.phue_label,
            type=self.type)


    @property
    def phue_unique(
        self,
    ) -> Optional[str]:
        """
        Return the unique identifier of group within the bridge.

        :returns: Unique identifier of group within the bridge.
        """

        source = self.phue_source

        if source is None:
            return None

        phid = source['id']

        assert isinstance(phid, str)

        return phid


    @property
    def phue_actual(
        self,
    ) -> Optional[str]:
        """
        Return the actual name for the group within the bridge.

        :returns: Actual name for the group within the bridge.
        """

        source = self.phue_source

        if source is None:
            return None

        metadata = source['metadata']
        name = metadata['name']

        assert isinstance(name, str)

        return name


    def scene_set(
        self,
        scene: 'HomieScene',
    ) -> None:
        """
        Update the current group to activate the provided scene.

        :param scene: Scene instance that will be used in match.
        """

        scene.scene_set(self)


    def homie_dumper(
        self,
    ) -> dict[str, Any]:
        """
        Return the content related to the project dumper script.

        :returns: Content related to the project dumper script.
        """

        params = (
            self.params.model_dump())

        phue_actual = self.phue_actual
        phue_unique = self.phue_unique
        phue_bridge = self.phue_bridge.name

        return {
            'name': self.name,
            'type': self.type,
            'present': bool(phue_unique),
            'phue_bridge': phue_bridge,
            'phue_unique': phue_unique,
            'phue_actual': phue_actual,
            'params': params}



"""
Functions and routines associated with Enasis Network Homie Automate.

This file is part of Enasis Network software eco-system. Distribution
is permitted, for more information consult the project license file.
"""



from json import loads
from pathlib import Path
from typing import TYPE_CHECKING

from encommon.utils import read_text
from encommon.utils import save_text

from respx import MockRouter

from . import SAMPLES
from ..params import WhatPhueButtonParams
from ..params import WhatPhueMotionParams
from ...conftest import config_factory
from ...conftest import homie_factory
from ...homie import HomieWhat
from ...homie import HomieWhatParams

if TYPE_CHECKING:
    from ...homie import Homie



def test_what_phue_motion(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    _events = loads(
        read_text(
            f'{SAMPLES}/events'
            '/source.json'))


    phue_motion = (
        WhatPhueMotionParams(
            device='jupiter_motion',
            sensor='motion1'))

    params = HomieWhatParams(
        phue_motion=phue_motion)

    what = HomieWhat(homie, params)

    assert what.match(_events[0])

    assert not what.match(_events[1])



def test_what_phue_motion_cover(
    tmp_path: Path,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param respx_mock: Object for mocking request operation.
    """

    _events = loads(
        read_text(
            f'{SAMPLES}/events'
            '/source.json'))

    config = (
        'phue_devices:\n'
        '  noexst:\n'
        '    bridge: jupiter\n'
        '    phid: noexst\n')


    Path.mkdir(
        tmp_path.joinpath('config'))

    save_text(
        f'{tmp_path}/config/test.yml',
        content=config)

    homie = homie_factory(
        config_factory(tmp_path),
        respx_mock)


    phue_motion = (
        WhatPhueMotionParams(
            device='noexst',
            sensor='motion1'))

    params = HomieWhatParams(
        phue_motion=phue_motion)

    what = HomieWhat(homie, params)

    assert not what.match(_events[0])

    assert not what.match(_events[1])


    phue_motion = (
        WhatPhueMotionParams(
            device='neptune_motion',
            sensor='motion1'))

    params = HomieWhatParams(
        phue_motion=phue_motion)

    what = HomieWhat(homie, params)

    assert not what.match(_events[0])

    assert not what.match(_events[1])


    assert not what.match({})



def test_what_phue_button(
    homie: 'Homie',
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param homie: Primary class instance for Homie Automate.
    """

    _events = loads(
        read_text(
            f'{SAMPLES}/events'
            '/source.json'))


    phue_button = (
        WhatPhueButtonParams(
            device='jupiter_button',
            sensor='button1',
            events=['initial_press']))

    params = HomieWhatParams(
        phue_button=phue_button)

    what = HomieWhat(homie, params)

    assert not what.match(_events[0])

    assert what.match(_events[1])



def test_what_phue_button_cover(
    tmp_path: Path,
    respx_mock: MockRouter,
) -> None:
    """
    Perform various tests associated with relevant routines.

    :param tmp_path: pytest object for temporal filesystem.
    :param respx_mock: Object for mocking request operation.
    """

    _events = loads(
        read_text(
            f'{SAMPLES}/events'
            '/source.json'))

    config = (
        'phue_devices:\n'
        '  noexst:\n'
        '    bridge: jupiter\n'
        '    phid: noexst\n')


    Path.mkdir(
        tmp_path.joinpath('config'))

    save_text(
        f'{tmp_path}/config/test.yml',
        content=config)

    homie = homie_factory(
        config_factory(tmp_path),
        respx_mock)


    phue_button = (
        WhatPhueButtonParams(
            device='noexst',
            sensor='button1',
            events=['initial_press']))

    params = HomieWhatParams(
        phue_button=phue_button)

    what = HomieWhat(homie, params)

    assert not what.match(_events[0])

    assert not what.match(_events[1])


    phue_button = (
        WhatPhueButtonParams(
            device='neptune_button',
            sensor='button1',
            events=['initial_press']))

    params = HomieWhatParams(
        phue_button=phue_button)

    what = HomieWhat(homie, params)

    assert not what.match(_events[0])

    assert not what.match(_events[1])


    phue_button = (
        WhatPhueButtonParams(
            device='jupiter_button',
            sensor='button2',
            events=['initial_press']))

    params = HomieWhatParams(
        phue_button=phue_button)

    what = HomieWhat(homie, params)

    assert not what.match(_events[0])

    assert not what.match(_events[1])


    assert not what.match({})
